#include "optimize.hh"

/*** This file contains all code pertaining to AST optimisation. It currently
     implements a simple optimisation called "constant folding". Most of the
     methods in this file are empty, or just relay optimize calls downward
     in the AST. If a more powerful AST optimization scheme were to be
     implemented, only methods in this file should need to be changed. ***/

static void	 optimize_binop(ast_binaryoperation *);
static void	 optimize_binrel(ast_binaryrelation *);


ast_optimizer *optimizer = new ast_optimizer();


/* The optimizer's interface method. Starts a recursive optimize call down
   the AST nodes, searching for binary operators with constant children. */
void ast_optimizer::do_optimize(ast_stmt_list *body)
{
    if (body != NULL) {
        body->optimize();
    }
}


/* Returns 1 if an AST expression is a subclass of ast_binaryoperation,
   ie, eligible for constant folding. */
bool ast_optimizer::is_binop(ast_expression *node)
{
    switch (node->tag) {
    case AST_ADD:
    case AST_SUB:
    case AST_OR:
    case AST_AND:
    case AST_MULT:
    case AST_DIVIDE:
    case AST_IDIV:
    case AST_MOD:
        return true;
    default:
        return false;
    }
}



/* We overload this method for the various ast_node subclasses that can
   appear in the AST. By use of virtual (dynamic) methods, we ensure that
   the correct method is invoked even if the pointers in the AST refer to
   one of the abstract classes such as ast_expression or ast_statement. */
void ast_node::optimize()
{
    fatal("Trying to optimize abstract class ast_node.");
}

void ast_statement::optimize()
{
    fatal("Trying to optimize abstract class ast_statement.");
}

void ast_expression::optimize()
{
    fatal("Trying to optimize abstract class ast_expression.");
}

void ast_lvalue::optimize()
{
    fatal("Trying to optimize abstract class ast_lvalue.");
}

void ast_binaryoperation::optimize()
{
    fatal("Trying to optimize abstract class ast_binaryoperation.");
}

void ast_binaryrelation::optimize()
{
    fatal("Trying to optimize abstract class ast_binaryrelation.");
}



/*** The optimize methods for the concrete AST classes. ***/

/* Optimize a statement list. */
void ast_stmt_list::optimize()
{
    if (preceding != NULL) {
        preceding->optimize();
    }
    if (last_stmt != NULL) {
        last_stmt->optimize();
    }
}


/* Optimize a list of expressions. */
void ast_expr_list::optimize()
{
    if (preceding)
        preceding->optimize();
    if (last_expr)
        last_expr = optimizer->fold_constants(last_expr);
}


/* Optimize an elsif list. */
void ast_elsif_list::optimize()
{
    if (preceding)
        preceding->optimize();
    if (last_elsif)
        last_elsif->optimize();
}


/* An identifier's value can change at run-time, so we can't perform
   constant folding optimization on it unless it is a constant.
   Thus we just do nothing here. It can be treated in the fold_constants()
   method, however. */
void ast_id::optimize()
{
	/* nil */
}

void ast_indexed::optimize()
{
    index->optimize();
}

/* This convenience method is used to apply constant folding to all
   binary operations. It returns either the resulting optimized node or the
   original node if no optimization could be performed. */
ast_expression *ast_optimizer::fold_constants(ast_expression *node)
{
	union val {
		long integer;
		double real;
	};
	ast_binaryoperation *binop;
	ast_expression *lop, *rop;
	union val lval, rval;

	node->optimize();

    if (node->tag == AST_ID) {
        ast_id *id = node->get_ast_id();
        if (sym_tab->get_symbol_tag(id->sym_p) == SYM_CONST) {
            constant_symbol *sym = sym_tab->get_symbol(id->sym_p)->get_constant_symbol();
            if (sym->type == integer_type)
                return new ast_integer(id->pos, sym->const_value.ival);
            return new ast_real(id->pos, sym->const_value.rval);
		}
    } else if (is_binop(node)) {

    	binop = node->get_ast_binaryoperation();
		lop = binop->left;
		rop = binop->right;

		if (lop->tag == AST_INTEGER)
			lval.integer = lop->get_ast_integer()->value;
		else if (binop->left->tag == AST_REAL)
			lval.real = lop->get_ast_real()->value;
		else
			return node;

		if (rop->tag == AST_INTEGER)
			rval.integer = rop->get_ast_integer()->value;
		else if (binop->right->tag == AST_REAL)
			rval.real = rop->get_ast_real()->value;
		else
			return node;

		switch(node->tag) {
		case AST_ADD:
			if (lop->tag == AST_INTEGER && rop->tag == AST_INTEGER)
				return new ast_integer(lop->pos, lval.integer + rval.integer);
			return new ast_real(lop->pos, lval.real + rval.real);
			break;
		case AST_AND:
			if (lop->tag == AST_INTEGER && rop->tag == AST_INTEGER)
				return new ast_integer(lop->pos, lval.integer && rval.integer);
			return new ast_real(lop->pos, lval.real && rval.real);
			break;
		case AST_DIVIDE:
			if (lop->tag == AST_REAL && rop->tag == AST_REAL)
				return new ast_integer(lop->pos, lval.real / rval.real);
			break;
		case AST_IDIV:
			if (lop->tag == AST_INTEGER && rop->tag == AST_INTEGER)
				return new ast_integer(lop->pos, lval.integer / rval.integer);
			break;
		case AST_MOD:
			if (lop->tag == AST_INTEGER && rop->tag == AST_INTEGER)
				return new ast_integer(lop->pos, lval.integer % rval.integer);
			break;
		case AST_MULT:
			if (lop->tag == AST_INTEGER && rop->tag == AST_INTEGER)
				return new ast_integer(lop->pos, lval.integer * rval.integer);
			return new ast_real(lop->pos, lval.real * rval.real);
			break;
		case AST_OR:
			if (lop->tag == AST_INTEGER && rop->tag == AST_INTEGER)
				return new ast_integer(lop->pos, lval.integer || rval.integer);
			return new ast_real(lop->pos, lval.real || rval.real);
			break;
		case AST_SUB:
			if (lop->tag == AST_INTEGER && rop->tag == AST_INTEGER)
				return new ast_integer(lop->pos, lval.integer - rval.integer);
			return new ast_real(lop->pos, lval.real - rval.real);
			break;
		default:
			break;
		}
    }
    return node;
}

static void
optimize_binop(ast_binaryoperation *binop)
{
	binop->left = optimizer->fold_constants(binop->left);
    binop->right = optimizer->fold_constants(binop->right);
}

static void
optimize_binrel(ast_binaryrelation *binrel)
{
	binrel->left = optimizer->fold_constants(binrel->left);
    binrel->right = optimizer->fold_constants(binrel->right);
}

/* All the binary operations should already have been detected in their parent
   nodes, so we don't need to do anything at all here. */
void ast_add::optimize()
{
    optimize_binop(this);
}

void ast_sub::optimize()
{
    optimize_binop(this);
}

void ast_mult::optimize()
{
    optimize_binop(this);
}

void ast_divide::optimize()
{
    optimize_binop(this);
}

void ast_or::optimize()
{
    optimize_binop(this);
}

void ast_and::optimize()
{
    optimize_binop(this);
}

void ast_idiv::optimize()
{
    optimize_binop(this);
}

void ast_mod::optimize()
{
    optimize_binop(this);
}

void ast_equal::optimize()
{
	optimize_binrel(this);
}

void ast_notequal::optimize()
{
    optimize_binrel(this);
}

void ast_lessthan::optimize()
{
    optimize_binrel(this);
}

void ast_greaterthan::optimize()
{
    optimize_binrel(this);
}



/*** The various classes derived from ast_statement. ***/

void ast_procedurecall::optimize()
{
     if (parameter_list)
        parameter_list->optimize();
}


void ast_assign::optimize()
{
    rhs->optimize();
    rhs = optimizer->fold_constants(rhs);
}


void ast_while::optimize()
{
    condition->optimize();
    condition = optimizer->fold_constants(condition);
    body->optimize();
}


void ast_if::optimize()
{
    condition->optimize();
    condition = optimizer->fold_constants(condition);
    body->optimize();
    if (elsif_list)
        elsif_list->optimize();
    if (else_body)
        else_body->optimize();
}


void ast_return::optimize()
{
    if (value) {
        value->optimize();
        value = optimizer->fold_constants(value);
    }
}


void ast_functioncall::optimize()
{
    if (parameter_list)
        parameter_list->optimize();
}

void ast_uminus::optimize()
{
    expr->optimize();
    expr = optimizer->fold_constants(expr);
}

void ast_not::optimize()
{
    expr->optimize();
    expr = optimizer->fold_constants(expr);
}


void ast_elsif::optimize()
{
    condition->optimize();
    condition = optimizer->fold_constants(condition);
    if (body)
        body->optimize();
}



void ast_integer::optimize()
{
    /* nothing to optimize here */
}

void ast_real::optimize()
{
    /* nothing to optimize here */
}

/* Note: See the comment in fold_constants() about casts and folding. */
void ast_cast::optimize()
{
    /* nothing to optimize here */
}



void ast_procedurehead::optimize()
{
    fatal("Trying to call ast_procedurehead::optimize()");
}


void ast_functionhead::optimize()
{
    fatal("Trying to call ast_functionhead::optimize()");
}
